#include "AS3935.hpp"

// C++ system level
// #include <functional>

// ESP32 specific
#include "esp_log.h"
#include "esp_err.h"
#include "driver/i2c.h"
#include "driver/gpio.h"

// project specific
#include <SpiderLib.hpp>

// qthing stuff
// #include <qthing>

// misc
// #include ""



static const char* TAG = "AS3935";


namespace AS3935 {


  AS3935::AS3935(u8 SDA, u8 SCL, u8 IRQ, u8 i2cAddr, u8 i2cChannel, u32 i2cFrequency) : i2cAddr(i2cAddr & I2C_ADDR_MASK), i2cChannel(i2cChannel) {

    // TODO: write ctor

    ESP_LOGI(TAG, "AS3935 test init…");




    i2c_config_t cfg;

    cfg.mode             = I2C_MODE_MASTER;
    cfg.sda_io_num       = (gpio_num_t)SDA;
    cfg.scl_io_num       = (gpio_num_t)SCL;
    cfg.sda_pullup_en    = GPIO_PULLUP_ENABLE;
    cfg.scl_pullup_en    = GPIO_PULLUP_ENABLE;
    cfg.master.clk_speed = i2cFrequency;

    i2c_param_config((i2c_port_t)i2cChannel, &cfg);

    esp_err_t result = i2c_driver_install((i2c_port_t)i2cChannel, cfg.mode, 0, 0, 0);

    ESP_LOGI(TAG, "I²C_driver_install(…): %s", result == ESP_OK ? "OK" : "FAILED!!");







  }




  bool AS3935::i2cWriteByte(u8 regAddr, u8 value) {

    i2c_cmd_handle_t cmd = i2c_cmd_link_create();
    i2c_master_start(cmd);
    i2c_master_write_byte(cmd, writeAddr(), true);
    i2c_master_write_byte(cmd, regAddr, true);
    i2c_master_write_byte(cmd, value, true);
    i2c_master_stop(cmd);
    esp_err_t err = i2c_master_cmd_begin((i2c_port_t)i2cChannel, cmd, 1000 / portTICK_PERIOD_MS);
    i2c_cmd_link_delete(cmd);

    return err == ESP_OK;
  }



  bool AS3935::i2cReadByte(u8 regAddr, u8* value) {

    if (true) {  // "normal" read routine


      i2c_cmd_handle_t cmd = i2c_cmd_link_create();

      ESP_ERROR_CHECK_WITHOUT_ABORT( i2c_master_start(cmd) );
      ESP_ERROR_CHECK_WITHOUT_ABORT( i2c_master_write_byte(cmd, readAddr(), true) );
      ESP_ERROR_CHECK_WITHOUT_ABORT( i2c_master_write_byte(cmd, regAddr, true) );

      ESP_ERROR_CHECK_WITHOUT_ABORT( i2c_master_read_byte(cmd, value, I2C_MASTER_LAST_NACK) );  // I2C_MASTER_ACK

      //i2c_master_write_byte(cmd, value, true);
      ESP_ERROR_CHECK_WITHOUT_ABORT( i2c_master_stop(cmd) );

      //ESP_ERROR_CHECK( i2c_master_cmd_begin((i2c_port_t)i2cChannel, cmd, 1000 / portTICK_PERIOD_MS) );
      esp_err_t err = i2c_master_cmd_begin((i2c_port_t)i2cChannel, cmd, 1000 / portTICK_PERIOD_MS);
      i2c_cmd_link_delete(cmd);
      ESP_ERROR_CHECK_WITHOUT_ABORT( err );

      return err == ESP_OK;


    } else {  // "0-byte-write-then-direction-change" read routine


      i2c_cmd_handle_t cmd = i2c_cmd_link_create();

      ESP_ERROR_CHECK_WITHOUT_ABORT( i2c_master_start(cmd) );

      ESP_ERROR_CHECK_WITHOUT_ABORT( i2c_master_write_byte(cmd, writeAddr(), true) );
      ESP_ERROR_CHECK_WITHOUT_ABORT( i2c_master_write_byte(cmd, regAddr, true) );

      ESP_ERROR_CHECK_WITHOUT_ABORT( i2c_master_start(cmd) );  // repeated start condition

      ESP_ERROR_CHECK_WITHOUT_ABORT( i2c_master_write_byte(cmd, readAddr(), true) );

      ESP_ERROR_CHECK_WITHOUT_ABORT( i2c_master_read_byte(cmd, value, I2C_MASTER_LAST_NACK) );  // I2C_MASTER_ACK  / or NACK ?   (manual says we have to NACK, but who's gonna sue us we have our data lol)

      ESP_ERROR_CHECK_WITHOUT_ABORT( i2c_master_stop(cmd) );

      //ESP_ERROR_CHECK( i2c_master_cmd_begin((i2c_port_t)i2cChannel, cmd, 1000 / portTICK_PERIOD_MS) );
      esp_err_t err = i2c_master_cmd_begin((i2c_port_t)i2cChannel, cmd, 1000 / portTICK_PERIOD_MS);
      i2c_cmd_link_delete(cmd);

      ESP_ERROR_CHECK_WITHOUT_ABORT( err );

      return err == ESP_OK;

    }
  }




  u8 AS3935::readAddr() const {
    return (i2cAddr << 1) & 1;
  }

  u8 AS3935::writeAddr() const {
    return i2cAddr << 1;
  }

}
