#pragma once

// C++ system level
// #include <functional>

// ESP32 specific
// #include "esp_log.h"

// project specific
#include <SpiderLib.hpp>

// qthing stuff
// #include <qthing>

// misc
// #include ""


namespace AS3935 {

  // 89's multiples are farther away from 500 kHz 
  // than the multiples of any other prime kHz clock between 50 kHz and 100 kHz
  constexpr u32 I2C_DEFAULT_FREQUENCY = 89000;

  constexpr u8 I2C_ADDR_MASK = 0x03;


  union R0x00 {

    struct __attribute__((packed)) {
      u8 PowerDown    : 1;
      u8 AFEGainBoost : 5;
      u8 _reserved_   : 2;
    };

    u8 value;

  };


  union R0x01 {
    struct __attribute__((packed)) {
      u8 WatchdogThreshold : 4;
      u8 NoiseFloorLevel   : 3;
      u8 _unused_          : 1;
    };

    u8 value;

  };


  union R0x02 {
    struct __attribute__((packed)) {
      u8 SpikeRejection  : 4;
      u8 MIN_NUM_LIGH    : 2;  // TODO: think of a good name without spelling it out completely!!
      u8 ClearStatistics : 1;
    };

    u8 value;

  };


  union R0x03 {
    struct __attribute__((packed)) {
      u8 Interrupt       : 4;
      u8 _reserved_      : 1;
      u8 MaskDisturber   : 1;
      u8 LCO_FrqDivision : 2;
    };

    u8 value;

  };


  union R0x04 {
    struct __attribute__((packed)) {
      u8 S_LIG_L : 8;  // TODO: think of a good name without spelling it out completely!!
    };

    u8 value;

  };


  union R0x05 {
    struct __attribute__((packed)) {
      u8 S_LIG_M : 8;  // TODO: think of a good name without spelling it out completely!!
    };

    u8 value;

  };


  union R0x06 {
    struct __attribute__((packed)) {
      u8 S_LIG_MM   : 5;  // TODO: think of a good name without spelling it out completely!!
      u8 _reserved_ : 3;
    };

    u8 value;

  };


  union R0x07 {
    struct __attribute__((packed)) {
      u8 Distance   : 6;
      u8 _reserved_ : 2;
    };

    u8 value;

  };


  union R0x08 {
    struct __attribute__((packed)) {
      u8 TuningCapacitor : 4;
      u8 Display_TRCO    : 1;
      u8 Display_SRCO    : 1;
      u8 Display_LCO     : 1;
    };

    u8 value;

  };


  union R0x3A {
    struct __attribute__((packed)) {
      u8 _reserved_            : 6;
      u8 TRCO_Calibration_NOK  : 1;
      u8 TRCO_Calibration_Done : 1;
    };

    u8 value;

  };


  union R0x3B {
    struct __attribute__((packed)) {
      u8 _reserved_            : 6;
      u8 SRCO_Calibration_NOK  : 1;
      u8 SRCO_Calibration_Done : 1;
    };

    u8 value;

  };










  class AS3935 {

    public:

      AS3935(u8 SDA, u8 SCL, u8 IRQ, u8 i2cAddr, u8 i2cChannel = 0, u32 i2cFrequency = I2C_DEFAULT_FREQUENCY);




      // Methods below are semi-public: From a hidden-box model perspective they should be protected
      // but from my POV they can be public as the user CANNOT misuse them in any way
      // TODO: elaborate on this…

      u8 readAddr () const;
      u8 writeAddr() const;


      bool i2cReadByte (u8 regAddr, u8* value);
      bool i2cWriteByte(u8 regAddr, u8  value);
    protected:

      const u8 i2cAddr;
      const u8 i2cChannel;


      //bool i2cReadByte (u8 regAddr, u8* value);
      //bool i2cWriteByte(u8 regAddr, u8  value);


  };

}
