#pragma once

// C++ system level
#include <functional>

// ESP32 specific
// #include "esp_log.h"

// project specific
#include <Types.hpp>

// qthing stuff
// #include <qthing>

// misc
// #include ""



namespace SpiderLib {


  class HasMQTT {
    public:

      using StringMsgHandler = std::function<void(str)>;

      bool mqttConnected() const;

      // API contract: must return true if msg was or will be published at a later moment.
      // API contract: returning true and storing msg until (re-)gaining connection is OK.
      // current implementation won't do this however and just return false if unconnected
      bool mqttPublish(const str topic, const str message, bool retained = false);

      // TODO: maybe think of an better API ??
      void mqttSubscribe(const str& topic, StringMsgHandler callback);



      // specializing on args count manually to avoid badly-designed va_args API 😞

      str deviceTopic() const;
      str deviceTopic(const str& subt0) const;
      str deviceTopic(const str& subt0, const str& subt1) const;
      str deviceTopic(const str& subt0, const str& subt1, const str& subt2) const;
      str deviceTopic(const str& subt0, const str& subt1, const str& subt2, const str& subt3) const;
      str deviceTopic(const str& subt0, const str& subt1, const str& subt2, const str& subt3, const str& subt4) const;
      str deviceTopic(const str& subt0, const str& subt1, const str& subt2, const str& subt3, const str& subt4, const str& subt5) const;
      str deviceTopic(const str& subt0, const str& subt1, const str& subt2, const str& subt3, const str& subt4, const str& subt5, const str& subt6) const;
      str deviceTopic(const str& subt0, const str& subt1, const str& subt2, const str& subt3, const str& subt4, const str& subt5, const str& subt6, const str& subt7) const;
      str deviceTopic(const str& subt0, const str& subt1, const str& subt2, const str& subt3, const str& subt4, const str& subt5, const str& subt6, const str& subt7, const str& subt8) const;
      str deviceTopic(const str& subt0, const str& subt1, const str& subt2, const str& subt3, const str& subt4, const str& subt5, const str& subt6, const str& subt7, const str& subt8, const str& subt9) const;

      str serviceTopic() const;
      str serviceTopic(const str& subt0) const;
      str serviceTopic(const str& subt0, const str& subt1) const;
      str serviceTopic(const str& subt0, const str& subt1, const str& subt2) const;
      str serviceTopic(const str& subt0, const str& subt1, const str& subt2, const str& subt3) const;
      str serviceTopic(const str& subt0, const str& subt1, const str& subt2, const str& subt3, const str& subt4) const;
      str serviceTopic(const str& subt0, const str& subt1, const str& subt2, const str& subt3, const str& subt4, const str& subt5) const;
      str serviceTopic(const str& subt0, const str& subt1, const str& subt2, const str& subt3, const str& subt4, const str& subt5, const str& subt6) const;
      str serviceTopic(const str& subt0, const str& subt1, const str& subt2, const str& subt3, const str& subt4, const str& subt5, const str& subt6, const str& subt7) const;
      str serviceTopic(const str& subt0, const str& subt1, const str& subt2, const str& subt3, const str& subt4, const str& subt5, const str& subt6, const str& subt7, const str& subt8) const;
      str serviceTopic(const str& subt0, const str& subt1, const str& subt2, const str& subt3, const str& subt4, const str& subt5, const str& subt6, const str& subt7, const str& subt8, const str& subt9) const;

      // specialization end

  };

}
